# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_get_init_creds_opt_set_pac_request(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        krb5_boolean req_pac,
    ) nogil


def get_init_creds_opt_set_pac_request(
    Context context not None,
    GetInitCredsOpt opt not None,
    req_pac: bool,
) -> None:
    cdef int value = 1 if req_pac else 0

    krb5_get_init_creds_opt_set_pac_request(context.raw, opt.raw, value)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_cc_dup(
        krb5_context context,
        krb5_ccache in_cc,
        krb5_ccache *out,
    ) nogil


def cc_dup(
    Context context not None,
    CCache cache not None,
) -> CCache:
    dup = CCache(context)
    cdef krb5_error_code err = 0

    err = krb5_cc_dup(context.raw, cache.raw, &dup.raw)
    if err:
        raise Krb5Error(context, err)

    return dup



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from libc.stdlib cimport free, malloc

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # Heimdal requires the context arg whereas MIT does not.
    """
    void krb5_get_init_creds_opt_set_canonicalize_generic(krb5_context context,
                                                          krb5_get_init_creds_opt *opt,
                                                          int canonicalize)
    {
    #if defined(HEIMDAL_XFREE)
        krb5_get_init_creds_opt_set_canonicalize(context, opt, (krb5_boolean)canonicalize);
    #else
        krb5_get_init_creds_opt_set_canonicalize(opt, canonicalize);
    #endif
    }
    """

    krb5_error_code krb5_get_init_creds_opt_alloc(
        krb5_context context,
        krb5_get_init_creds_opt **opt,
    ) nogil

    void krb5_get_init_creds_opt_set_anonymous(
        krb5_get_init_creds_opt *opt,
        int anonymous,
    ) nogil

    void krb5_get_init_creds_opt_free(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
    ) nogil

    void krb5_get_init_creds_opt_set_canonicalize_generic(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        int canonicalize,
    ) nogil

    void krb5_get_init_creds_opt_set_etype_list(
        krb5_get_init_creds_opt *opt,
        krb5_enctype *etype_list,
        int etype_list_length,
    ) nogil

    void krb5_get_init_creds_opt_set_forwardable(
        krb5_get_init_creds_opt *opt,
        int forwardable,
    ) nogil

    void krb5_get_init_creds_opt_set_proxiable(
        krb5_get_init_creds_opt *opt,
        int proxiable,
    ) nogil

    void krb5_get_init_creds_opt_set_renew_life(
        krb5_get_init_creds_opt *opt,
        krb5_deltat renew_life,
    ) nogil

    void krb5_get_init_creds_opt_set_salt(
        krb5_get_init_creds_opt *opt,
        krb5_data *salt,
    ) nogil

    void krb5_get_init_creds_opt_set_tkt_life(
        krb5_get_init_creds_opt *opt,
        krb5_deltat tkt_life,
    ) nogil


cdef class GetInitCredsOpt:
    # cdef Context ctx
    # cdef krb5_get_init_creds_opt *raw

    def __cinit__(GetInitCredsOpt self, Context context):
        self.ctx = context
        self.raw = NULL

    def __dealloc__(GetInitCredsOpt self):
        if self.raw:
            krb5_get_init_creds_opt_free(self.ctx.raw, self.raw)
            self.raw = NULL

    def __str__(GetInitCredsOpt self):
        return "GetInitCredsOpt"


def get_init_creds_opt_alloc(
    Context context not None,
) -> GetInitCredsOpt:
    opt = GetInitCredsOpt(context)
    cdef krb5_error_code err = 0

    err = krb5_get_init_creds_opt_alloc(context.raw, &opt.raw)
    if err:
        raise Krb5Error(context, err)

    return opt


def get_init_creds_opt_set_anonymous(
    GetInitCredsOpt opt not None,
    anonymous: bool,
) -> None:
    cdef int value = 1 if anonymous else 0

    krb5_get_init_creds_opt_set_anonymous(opt.raw, value)


def get_init_creds_opt_set_canonicalize(
    GetInitCredsOpt opt not None,
    canonicalize: bool,
) -> None:
    cdef int value = 1 if canonicalize else 0

    krb5_get_init_creds_opt_set_canonicalize_generic(opt.ctx.raw, opt.raw, value)


def get_init_creds_opt_set_etype_list(
    GetInitCredsOpt opt not None,
    etypes: typing.Iterable[int],
) -> None:
    tmp = list(etypes)
    cdef krb5_enctype *buffer = <krb5_enctype *>malloc(len(etypes) * sizeof(krb5_enctype))
    if not buffer:
        raise MemoryError()

    for idx, e in enumerate(tmp):
        buffer[idx] = tmp[idx]

    krb5_get_init_creds_opt_set_etype_list(opt.raw, buffer, len(etypes))


def get_init_creds_opt_set_forwardable(
    GetInitCredsOpt opt not None,
    forwardable: bool,
) -> None:
    cdef int value = 1 if forwardable else 0

    krb5_get_init_creds_opt_set_forwardable(opt.raw, value)


def get_init_creds_opt_set_proxiable(
    GetInitCredsOpt opt not None,
    proxiable: bool,
) -> None:
    cdef int value = 1 if proxiable else 0

    krb5_get_init_creds_opt_set_proxiable(opt.raw, value)


def get_init_creds_opt_set_renew_life(
    GetInitCredsOpt opt not None,
    krb5_deltat renew_life,
) -> None:
    krb5_get_init_creds_opt_set_renew_life(opt.raw, renew_life)


def get_init_creds_opt_set_salt(
    GetInitCredsOpt opt not None,
    const unsigned char[:] salt not None,
) -> None:
    if not len(salt):
        raise ValueError("salt cannot be an empty byte string")

    cdef krb5_data buffer
    pykrb5_set_krb5_data(&buffer, len(salt), <char *>&salt[0])
    krb5_get_init_creds_opt_set_salt(opt.raw, &buffer)


def get_init_creds_opt_set_tkt_life(
    GetInitCredsOpt opt not None,
    krb5_deltat tkt_life,
) -> None:
    krb5_get_init_creds_opt_set_tkt_life(opt.raw, tkt_life)



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_get_init_creds_opt_set_in_ccache(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        krb5_ccache ccache,
    ) nogil


def get_init_creds_opt_set_in_ccache(
    Context context not None,
    GetInitCredsOpt opt not None,
    CCache ccache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_get_init_creds_opt_set_in_ccache(context.raw, opt.raw, ccache.raw)
    if err:
        raise Krb5Error(context, err)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdint cimport int32_t, uint8_t, uint32_t


cdef extern from "python_krb5.h":
    """
    // The Heimdal and MIT krb5 libraries have different implementations of the krb5_data struct.
    // MIT uses a struct with a member magic of -1760647422L
    // Heimdal uses a struct without a magic
    void pykrb5_init_krb5_data(
        krb5_data *data
    )
    {
#if defined(HEIMDAL_XFREE)
        krb5_data_zero(data);
#else
        data->magic = KV5M_DATA;
        data->length = 0;
        data->data = NULL;
#endif
    }

    // The structures are different so cannot be explicitly defined in Cython. Use inline C to set the structs elements
    // by name.
    void pykrb5_set_krb5_data(
        krb5_data *data,
        size_t length,
        char *value
    )
    {
#if !defined(HEIMDAL_XFREE)
        data->magic = KV5M_DATA;
#endif
        data->length = length;
        data->data = value;
    }

    void pykrb5_get_krb5_data(
        krb5_data *data,
        size_t *length,
        char **value
    )
    {
        if (length) *length = data->length;
        if (value) *value = data->data;
    }

    void pykrb5_free_data_contents(
        krb5_context context,
        krb5_data *val
    )
    {
#if defined(HEIMDAL_XFREE)
        krb5_data_free(val);
#else
        krb5_free_data_contents(context, val);
#endif
    }
    """

    ctypedef int32_t krb5_int32
    ctypedef krb5_int32 krb5_error_code
    ctypedef krb5_int32 krb5_deltat
    ctypedef krb5_int32 krb5_enctype
    ctypedef uint8_t krb5_octet
    ctypedef krb5_int32 krb5_timestamp
    ctypedef unsigned int krb5_boolean
    ctypedef unsigned int krb5_kvno

    ctypedef void *krb5_pointer;
    ctypedef krb5_pointer krb5_cc_cursor
    ctypedef krb5_pointer krb5_kt_cursor;

    cdef struct _krb5_context:
        pass
    ctypedef _krb5_context *krb5_context

    cdef struct _krb5_cccol_cursor:
        pass
    ctypedef _krb5_cccol_cursor *krb5_cccol_cursor

    ctypedef struct krb5_principal_data:
        pass
    ctypedef krb5_principal_data *krb5_principal
    ctypedef const krb5_principal_data *krb5_const_principal

    cdef struct _krb5_ccache:
        pass
    ctypedef _krb5_ccache *krb5_ccache

    cdef struct _krb5_get_init_creds_opt:
        pass
    ctypedef _krb5_get_init_creds_opt krb5_get_init_creds_opt

    cdef struct _krb5_init_creds_context:
        pass
    ctypedef _krb5_init_creds_context *krb5_init_creds_context

    cdef struct _krb5_keyblock:
        pass
    ctypedef _krb5_keyblock krb5_keyblock

    cdef struct _krb5_data:
        pass
    ctypedef _krb5_data krb5_data

    cdef struct _krb5_kt:
        pass
    ctypedef _krb5_kt *krb5_keytab

    cdef struct krb5_keytab_entry_st:
        pass
    ctypedef krb5_keytab_entry_st krb5_keytab_entry

    cdef struct _krb5_creds:
        pass
    ctypedef _krb5_creds krb5_creds

    cdef struct _krb5_prompt:
        char *prompt
        int hidden
        krb5_data *reply
    ctypedef _krb5_prompt krb5_prompt

    cdef struct _pykrb5_ticket_times:
        krb5_timestamp authtime
        krb5_timestamp starttime
        krb5_timestamp endtime
        krb5_timestamp renew_till
    ctypedef _pykrb5_ticket_times pykrb5_ticket_times

    ctypedef krb5_error_code(*krb5_prompter_fct)(
        krb5_context context,
        void *data,
        const char *name,
        const char *banner,
        int num_prompts,
        krb5_prompt prompts[],
    )

    void pykrb5_init_krb5_data(
        krb5_data *data,
    ) nogil

    void pykrb5_set_krb5_data(
        krb5_data *data,
        size_t length,
        char *value,
    ) nogil

    void pykrb5_get_krb5_data(
        krb5_data *data,
        size_t *length,
        char **value,
    ) nogil

    void pykrb5_free_data_contents(
        krb5_context context,
        krb5_data *val,
    ) nogil



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._exceptions import Krb5Error

from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # krb5_free_default_realm is optionally exported in Heimdal (not at all on macOS) - use krb5_xfree instead
    """
    void krb5_free_default_realm_generic(krb5_context context, char *realm)
    {
    #if defined(HEIMDAL_XFREE)
        krb5_xfree(realm);
    #else
        krb5_free_default_realm(context, realm);
    #endif
    }
    """

    void krb5_free_context(
        krb5_context context,
    ) nogil

    krb5_error_code krb5_init_context(
        krb5_context *context,
    ) nogil

    # See inline C code
    void krb5_free_default_realm_generic(
        krb5_context context,
        char *realm,
    ) nogil

    krb5_error_code krb5_get_default_realm(
        krb5_context context,
        char **realm,
    ) nogil

    krb5_error_code krb5_set_default_realm(
        krb5_context context,
        const char *realm
    ) nogil

    krb5_error_code krb5_timeofday(
        krb5_context context,
        krb5_timestamp *seconds
    ) nogil

    krb5_error_code krb5_us_timeofday(
        krb5_context context,
        krb5_timestamp *seconds,
        int32_t *microseconds
    ) nogil

    krb5_error_code krb5_set_real_time(
        krb5_context context,
        krb5_timestamp seconds,
        int32_t microseconds
    ) nogil


cdef class Context:
    # cdef krb5_context raw

    def __cinit__(Context self):
        self.raw = NULL

    def __dealloc__(Context self):
        if self.raw:
            krb5_free_context(self.raw)
            self.raw = NULL

    def __str__(Context self):
        return "Krb5Context"


def init_context() -> Context:
    cdef krb5_error_code = 0
    context = Context()

    err = krb5_init_context(&context.raw)
    if err:
        raise Krb5Error(context, err)

    return context


def get_default_realm(
    Context context not None,
) -> bytes:
    cdef krb5_error_code = 0
    cdef char *realm = NULL

    err = krb5_get_default_realm(context.raw, &realm)
    if err:
        raise Krb5Error(context, err)

    try:
        return <bytes>realm
    finally:
        krb5_free_default_realm_generic(context.raw, realm)


def set_default_realm(
    Context context not None,
    const unsigned char[:] realm,
) -> None:
    cdef krb5_error_code = 0
    cdef const char *realm_ptr = NULL
    if realm is not None and len(realm):
        realm_ptr = <const char*>&realm[0]

    err = krb5_set_default_realm(context.raw, realm_ptr)
    if err:
        raise Krb5Error(context, err)


def timeofday(
    Context context not None,
) -> int:
    cdef krb5_error_code = 0

    cdef krb5_timestamp seconds
    err = krb5_timeofday(context.raw, &seconds)
    if err:
        raise Krb5Error(context, err)

    return seconds


def us_timeofday(
    Context context not None,
) -> typing.Tuple[int, int]:
    cdef krb5_error_code = 0

    cdef krb5_timestamp seconds
    cdef int32_t microseconds
    err = krb5_us_timeofday(context.raw, &seconds, &microseconds)
    if err:
        raise Krb5Error(context, err)

    return seconds, microseconds


def set_real_time(
    Context context not None,
    int seconds,
    int microseconds,
) -> None:
    cdef krb5_error_code = 0

    err = krb5_set_real_time(context.raw, seconds, microseconds)
    if err:
        raise Krb5Error(context, err)



# Copyright: (c) 2024 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum
import struct
import typing

FORMAT = "!HIIIQQ"


class ADPolicyInfoProp(enum.IntFlag):
    COMPLEX = 0x00000001
    NO_ANON_CHANGEv = 0x00000002
    NO_CLEAR_CHANGE = 0x00000004
    LOCKOUT_ADMINS = 0x00000008
    STORE_CLEARTEXT = 0x00000010
    REFUSE_CHANGE = 0x00000020


class ADPolicyInfo(typing.NamedTuple):
    """The structure containing the reasons for failed password change attempt.
    Should be used to inform the end user how to meet the policy requirements.
    This is specific to Active Directory and is returned as the
    `server_response` by :meth:`set_password()` and
    :meth:`set_password_using_ccache()`.

    When using MIT library, this structure may be encoded back to bytes and
    passed to :meth:`chpw_message()` to obtain a human readable response.
    With Heimdal, it is required to provide a custom implementation based
    on the known fields below.

    The structure contains the following fields:\n
    - `properties` - Password policy bit flags (see below)
    - `min_length` - Minimal password length
    - `history`    - Number of passwords that this system remembers
    - `max_age`    - Maximum password age in 100 nanosecond units
    - `min_age`    - Minimum password age in 100 nanosecond units

    The only known property flag is `COMPLEX` which means that the password must
    meet certain character variety and not contain the user's name.
    To convert `max_age` and `min_age` to seconds, divide them by 10,000,000.
    """

    properties: ADPolicyInfoProp
    min_length: int
    history: int
    max_age: int
    min_age: int

    @classmethod
    def from_bytes(cls, data: bytes) -> ADPolicyInfo:
        """Decode AD policy result from byte string

        Args:
            data: Serialized AD policy `server_response`

        Returns:
            ADPolicyInfo: Decoded AD policy result strcture

        Raises:
            ValueError: Invalid data length or wrong signature
        """
        if len(data) != struct.calcsize(FORMAT):
            raise ValueError("Invalid data length")
        signature, min_length, history, flags, max_age, min_age = struct.unpack(FORMAT, data)
        if signature != 0x0000:
            raise ValueError("Invalid signature")
        return cls(
            min_length=min_length,
            history=history,
            max_age=max_age,
            min_age=min_age,
            properties=ADPolicyInfoProp(flags),
        )



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *
from krb5._kt cimport KeyTab


cdef extern from "python_krb5.h":
    krb5_error_code krb5_kt_get_full_name(
        krb5_context context,
        krb5_keytab keytab,
        char **str,
    ) nogil

    krb5_error_code krb5_xfree(
        void *ptr,
    ) nogil

    krb5_error_code KRB5_KT_PREFIX_MAX_LEN


def kt_get_full_name(
    Context context not None,
    KeyTab keytab not None,
) -> bytes:
    cdef krb5_error_code err = 0
    cdef char *str = NULL

    err = krb5_kt_get_full_name(context.raw, keytab.raw, &str)
    if err:
        raise Krb5Error(context, err)

    try:
        return <bytes>str
    finally:
        krb5_xfree(str)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

class Context:
    """Kerberos Library Context

    This class represents a library context object.
    """

def init_context() -> Context:
    """Create a krb5 library context.

    Creates a krb5 library context.

    Returns:
        Context: The opened krb5 library context.
    """

def get_default_realm(
    context: Context,
) -> bytes:
    """Get default realm for the specified context.

    Returns the default realm of the context passed in.

    Args:
        context: Krb5 context.

    Returns:
        bytes: The default realm of the context.
    """

def set_default_realm(
    context: Context,
    realm: typing.Optional[bytes],
) -> None:
    """Override the default realm for the specified context.

    Sets the default realm of the passed in context to the value specified. Use
    ``None`` to clear out the existing explicit setting.

    Args:
        context: Krb5 context.
        realm: The realm to set as the default realm.
    """

def timeofday(
    context: Context,
) -> int:
    """Return the adjusted time.

    Return the adjusted time.

    Args:
        context: Krb5 context.

    Returns:
        The current time as seen by the KDC in seconds.
    """

def us_timeofday(
    context: Context,
) -> typing.Tuple[int, int]:
    """Return the adjusted time with microseconds.

    Return the adjusted time with microseconds.

    Args:
        context: Krb5 context.

    Returns:
        The current time as seen by the KDC in seconds and microseconds.
    """

def set_real_time(
    context: Context,
    seconds: int,
    microseconds: int,
) -> None:
    """Set the time offset to the difference between the system time and the specified time.

    Set the time offset of the context to the difference between the system time
    and the specified time.

    Args:
        context: Krb5 context.
        seconds: The seconds of the current time as seen by the KDC.
        microseconds: The microseconds of the current time as seen by the KDC or -1.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import collections
import enum
import typing

from krb5._exceptions import Krb5Error
from krb5._keyblock import copy_keyblock
from krb5._principal import copy_principal

from libc.stdlib cimport calloc, free

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._keyblock cimport KeyBlock
from krb5._krb5_types cimport *
from krb5._kt cimport KeyTab
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    """
    void pykrb5_creds_get(
        krb5_creds *creds,
        krb5_principal *client,
        krb5_principal *server,
        krb5_keyblock **keyblock,
        pykrb5_ticket_times *times,
        uint32_t *ticket_flags_raw,
        uint32_t *ticket_flags,
        // krb5_address ***addresses,
        krb5_data *ticket,
        krb5_data *second_ticket
        // krb5_authdata ***authdata,
    )
    {
        if (client != NULL) *client = creds->client;
        if (server != NULL) *server = creds->server;
#if defined(HEIMDAL_XFREE)
        if (keyblock != NULL) *keyblock = &creds->session;
#else
        if (keyblock != NULL) *keyblock = &creds->keyblock;
#endif
        if (times != NULL) *times = creds->times;
#if defined(HEIMDAL_XFREE)
        if (ticket_flags_raw != NULL) *ticket_flags_raw = creds->flags.i;
        if (ticket_flags != NULL) *ticket_flags = creds->flags.i;
#else
        if (ticket_flags_raw != NULL) *ticket_flags_raw = creds->ticket_flags;
        if (ticket_flags != NULL) {
            *ticket_flags = 0;
            // Reverse the order of the bits to get the first bit in the
            // lowermost bit instead of in the uppermost bit.
            for (int i = 0; i < 32; i++) {
                if (creds->ticket_flags & (1 << (31 - i))) *ticket_flags |= (1 << i);
            }
        }
#endif
        // if (addresses != NULL) *addresses = creds->addresses;
        if (ticket != NULL) *ticket = creds->ticket;
        if (second_ticket != NULL) *second_ticket = creds->second_ticket;
        // if (authdata != NULL) *authdata = creds->authdata;
    }
    """

    void pykrb5_creds_get(
        krb5_creds *creds,
        krb5_principal *client,
        krb5_principal *server,
        krb5_keyblock **keyblock,
        pykrb5_ticket_times *times,
        uint32_t *ticket_flags_raw,
        uint32_t *ticket_flags,
        # krb5_address ***addresses,
        krb5_data *ticket,
        krb5_data *second_ticket,
        # krb5_authdata ***authdata,
    ) nogil

    void krb5_free_creds(
        krb5_context context,
        krb5_creds *val,
    ) nogil

    void krb5_free_cred_contents(
        krb5_context context,
        krb5_creds *val,
    ) nogil

    krb5_error_code krb5_init_creds_get(
        krb5_context context,
        krb5_init_creds_context ctx,
    ) nogil

    krb5_error_code krb5_init_creds_get_creds(
        krb5_context context,
        krb5_init_creds_context ctx,
        krb5_creds *creds,
    ) nogil

    krb5_error_code krb5_init_creds_init(
        krb5_context context,
        krb5_principal client,
        krb5_prompter_fct prompert,
        void *data,
        krb5_deltat start_time,
        krb5_get_init_creds_opt *options,
        krb5_init_creds_context *ctx,
    ) nogil

    krb5_error_code krb5_get_init_creds_keytab(
        krb5_context context,
        krb5_creds *creds,
        krb5_principal client,
        krb5_keytab arg_keytab,
        krb5_deltat start_time,
        const char *in_tkt_service,
        krb5_get_init_creds_opt *k5_gic_options,
    ) nogil

    void krb5_init_creds_free(
        krb5_context context,
        krb5_init_creds_context ctx,
    ) nogil

    krb5_error_code krb5_get_init_creds_password(
        krb5_context context,
        krb5_creds *creds,
        krb5_principal client,
        const char *password,
        krb5_prompter_fct prompter,
        void *data,
        krb5_deltat start_time,
        const char *in_tkt_service,
        krb5_get_init_creds_opt *k5_gic_options,
    ) nogil

    krb5_error_code krb5_init_creds_set_keytab(
        krb5_context context,
        krb5_init_creds_context ctx,
        krb5_keytab keytab,
    ) nogil

    krb5_error_code krb5_init_creds_set_password(
        krb5_context context,
        krb5_init_creds_context ctx,
        const char *password,
    ) nogil

    krb5_error_code krb5_get_renewed_creds(
        krb5_context context,
        krb5_creds *creds,
        krb5_principal client,
        krb5_ccache ccache,
        const char *in_tkt_service,
    ) nogil


class TicketFlags(enum.IntFlag):
    # https://github.com/krb5/krb5-assignments/blob/master/ticket-flags
    reserved = 1 << 0
    forwardable = 1 << 1
    forwarded = 1 << 2
    proxiable = 1 << 3
    proxy = 1 << 4
    may_postdate = 1 << 5
    postdated = 1 << 6
    invalid = 1 << 7
    renewable = 1 << 8
    initial = 1 << 9
    pre_authent = 1 << 10
    hw_authent = 1 << 11
    transited_policy_checked = 1 << 12
    ok_as_delegate = 1 << 13
    enc_pa_rep = 1 << 15
    anonymous = 1 << 16

    # This is to prevent python >= 3.11 from clearing unknown flags when doing:
    # flags = flags & ~TicketFlags.forwarded
    # (Under python 3.11, ~TicketFlags.forwarded will contain only known flags.)
    _all_flags = (1 << 32) - 1


cdef class Creds:
    # cdef Context ctx
    # cdef int free_contents
    # cdef krb5_creds* _raw
    # cdef int _free_raw

    def __cinit__(Creds self, Context context):
        self.ctx = context
        self.free_contents = 0
        self._raw = NULL
        self._free_raw = 0

    def __dealloc__(Creds self):
        if not self._raw:
            return

        if self._free_raw:
            if self.free_contents:
                krb5_free_cred_contents(self.ctx.raw, self._raw)
                self.free_contents = 0
            free(self._raw)

        else:
            krb5_free_creds(self.ctx.raw, self._raw)

        self._raw = NULL
        self._free_raw = 0

    def __str__(Creds self) -> str:
        return "Creds"

    cdef void* set_raw_from_lib(Creds self, krb5_creds* raw):
        # This is called when the krb5_creds* was allocated by the krb5 lib.
        # We set the internal state so it calls krb5_free_creds properly on
        # deallocation.
        if self._raw:
            # This should never occur but it's here just in case
            raise ValueError("Cannot set raw pointer on already initialized Creds")

        self._raw = raw
        self._free_raw = 0

    cdef krb5_creds *get_pointer(Creds self):
        # If this is being called without _raw being set we want to manage
        # the krb5_creds object in this class.
        if not self._raw:
            self._raw = <krb5_creds*>calloc(sizeof(krb5_creds), 1)
            if not self._raw:
                raise MemoryError("Cannot calloc krb5_creds")
            self._free_raw = 1

        return self._raw

    @property
    def client(Creds self) -> Principal:
        princ = Principal(self.ctx, 0, needs_free=0)
        pykrb5_creds_get(self.get_pointer(), &princ.raw, NULL, NULL, NULL, NULL, NULL, NULL, NULL)

        # Create a copy of the principal to make sure the returned value
        # remains valid even if the Creds object is destroyed
        princ_copy = copy_principal(self.ctx, princ)

        return princ_copy

    @property
    def server(Creds self) -> Principal:
        princ = Principal(self.ctx, 0, needs_free=0)
        pykrb5_creds_get(self.get_pointer(), NULL, &princ.raw, NULL, NULL, NULL, NULL, NULL, NULL)

        # Create a copy of the principal to make sure the returned value
        # remains valid even if the Creds object is destroyed
        princ_copy = copy_principal(self.ctx, princ)

        return princ_copy

    @property
    def keyblock(Creds self) -> KeyBlock:
        kb = KeyBlock(self.ctx, needs_free=0)
        pykrb5_creds_get(self.get_pointer(), NULL, NULL, &kb.raw, NULL, NULL, NULL, NULL, NULL)

        # Create a copy of the keyblock to make sure the returned value
        # remains valid even if the Creds object is destroyed
        kb_copy = copy_keyblock(self.ctx, kb)

        return kb_copy

    @property
    def times(Creds self) -> TicketTimes:
        cdef pykrb5_ticket_times times
        pykrb5_creds_get(self.get_pointer(), NULL, NULL, NULL, &times, NULL, NULL, NULL, NULL)

        return TicketTimes(times.authtime, times.starttime, times.endtime, times.renew_till)

    @property
    def ticket_flags_raw(Creds self) -> int:
        cdef uint32_t flags_raw
        pykrb5_creds_get(self.get_pointer(), NULL, NULL, NULL, NULL, &flags_raw, NULL, NULL, NULL)

        return flags_raw

    @property
    def ticket_flags(Creds self) -> TicketFlags:
        cdef uint32_t flags
        pykrb5_creds_get(self.get_pointer(), NULL, NULL, NULL, NULL, NULL, &flags, NULL, NULL)

        return TicketFlags(flags)

    @property
    def ticket(Creds self) -> bytes:
        cdef krb5_data ticket
        pykrb5_creds_get(self.get_pointer(), NULL, NULL, NULL, NULL, NULL, NULL, &ticket, NULL)

        cdef size_t length
        cdef char *value
        pykrb5_get_krb5_data(&ticket, &length, &value)

        if length == 0:
            return b""
        else:
            return value[:length]

    @property
    def second_ticket(Creds self) -> bytes:
        cdef krb5_data second_ticket
        pykrb5_creds_get(self.get_pointer(), NULL, NULL, NULL, NULL, NULL, NULL, NULL, &second_ticket)

        cdef size_t length
        cdef char *value
        pykrb5_get_krb5_data(&second_ticket, &length, &value)

        if length == 0:
            return b""
        else:
            return value[:length]

cdef class InitCredsContext:
    # cdef Context ctx
    # cdef krb5_init_creds_context raw

    def __cinit__(InitCredsContext self, Context context):
        self.ctx = context
        self.raw = NULL

    def __dealloc__(InitCredsContext self):
        if self.raw:
            krb5_init_creds_free(self.ctx.raw, self.raw)
            self.raw = NULL

    def __str__(InitCredsContext self) -> str:
        return "InitCredsContext"


cdef class Krb5Prompt:
    def init(
        self,
        name: typing.Optional[bytes],
        banner: typing.Optional[bytes],
        num_prompts: int,
    ) -> None:
        pass

    def prompt(
        self,
        msg: bytes,
        hidden: bool,
    ) -> bytes:
        raise NotImplementedError()


cdef krb5_error_code prompt_callback(
    krb5_context context,
    void *data,
    const char *name,
    const char *banner,
    int num_prompts,
    krb5_prompt *prompts,
) noexcept with gil:
    try:
        prompter = <Krb5Prompt>data

        py_name = None if name == NULL else <bytes>name
        py_banner = None if banner == NULL else <bytes>banner
        prompter.init(py_name, py_banner, num_prompts)

        replies = []
        for prompt in prompts[:num_prompts]:
            msg = <bytes>prompt.prompt
            hidden = prompt.hidden != 0

            reply = prompter.prompt(msg, hidden)
            if not isinstance(reply, bytes):
                return 1

            replies.append(reply)

        for idx, reply in enumerate(replies):
            pykrb5_set_krb5_data(prompts[idx].reply, len(reply), <char *>reply)

        return 0

    except Exception:
        return 1


def get_init_creds_keytab(
    Context context not None,
    Principal client not None,
    GetInitCredsOpt k5_gic_options not None,
    KeyTab keytab not None,
    int start_time = 0,
    const unsigned char[:] in_tkt_service = None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0
    cdef krb5_creds* raw_creds = creds.get_pointer()

    cdef const char *in_tkt_service_ptr = NULL
    if in_tkt_service is not None and len(in_tkt_service):
        in_tkt_service_ptr = <const char*>&in_tkt_service[0]

    with nogil:
        err = krb5_get_init_creds_keytab(
            context.raw,
            raw_creds,
            client.raw,
            keytab.raw,
            start_time,
            in_tkt_service_ptr,
            k5_gic_options.raw,
        )

    if err:
        raise Krb5Error(context, err)

    creds.free_contents = 1

    return creds


def get_init_creds_password(
    Context context not None,
    Principal client not None,
    GetInitCredsOpt k5_gic_options not None,
    const unsigned char[:] password = None,
    int start_time = 0,
    const unsigned char[:] in_tkt_service = None,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0
    cdef krb5_creds* raw_creds = creds.get_pointer()

    cdef const char *password_ptr = NULL
    if password is not None and len(password):
        password_ptr = <const char*>&password[0]

    cdef krb5_prompter_fct callback = NULL
    cdef void *prompt_data = NULL
    if prompter is not None:
        callback = prompt_callback
        prompt_data = <void*>prompter

    cdef const char *in_tkt_service_ptr = NULL
    if in_tkt_service is not None and len(in_tkt_service):
        in_tkt_service_ptr = <const char*>&in_tkt_service[0]

    with nogil:
        err = krb5_get_init_creds_password(
            context.raw,
            raw_creds,
            client.raw,
            password_ptr,
            callback,
            prompt_data,
            start_time,
            in_tkt_service_ptr,
            k5_gic_options.raw,
        )

    if err:
        raise Krb5Error(context, err)

    creds.free_contents = 1

    return creds


def init_creds_get(
    Context context not None,
    InitCredsContext ctx not None,
) -> None:
    cdef krb5_error_code err = 0

    with nogil:
        err = krb5_init_creds_get(context.raw, ctx.raw)

    if err:
        raise Krb5Error(context, err)


def init_creds_get_creds(
    Context context not None,
    InitCredsContext ctx not None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    err = krb5_init_creds_get_creds(context.raw, ctx.raw, creds.get_pointer())
    if err:
        raise Krb5Error(context, err)

    creds.free_contents = 1
    return creds


def init_creds_init(
    Context context not None,
    Principal client not None,
    GetInitCredsOpt k5_gic_options = None,
    int start_time = 0,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> InitCredsContext:
    creds_ctx = InitCredsContext(context)
    cdef krb5_error_code err = 0

    cdef krb5_get_init_creds_opt *options = NULL
    if k5_gic_options:
        options = k5_gic_options.raw

    cdef krb5_prompter_fct callback = NULL
    cdef void *prompt_data = NULL
    if prompter is not None:
        callback = prompt_callback
        prompt_data = <void*>prompter

    with nogil:
        err = krb5_init_creds_init(
            context.raw,
            client.raw,
            callback,
            prompt_data,
            start_time,
            options,
            &creds_ctx.raw
        )

    if err:
        raise Krb5Error(context, err)

    return creds_ctx


def init_creds_set_keytab(
    Context context not None,
    InitCredsContext ctx not None,
    KeyTab keytab not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_init_creds_set_keytab(context.raw, ctx.raw, keytab.raw)
    if err:
        raise Krb5Error(context, err)


def init_creds_set_password(
    Context context not None,
    InitCredsContext ctx not None,
    const unsigned char[:] password,
) -> None:
    cdef krb5_error_code err = 0

    cdef const char *password_ptr = NULL
    if password is not None and len(password):
        password_ptr = <const char*>&password[0]
    else:
        raise ValueError("password must be set")

    err = krb5_init_creds_set_password(context.raw, ctx.raw, password_ptr)
    if err:
        raise Krb5Error(context, err)

def get_renewed_creds(
    Context context not None,
    Principal client not None,
    CCache ccache not None,
    const unsigned char[:] in_tkt_service = None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    cdef const char *in_tkt_service_ptr = NULL
    if in_tkt_service is not None and len(in_tkt_service):
        in_tkt_service_ptr = <const char*>&in_tkt_service[0]

    err = krb5_get_renewed_creds(
        context.raw,
        creds.get_pointer(),
        client.raw,
        ccache.raw,
        in_tkt_service_ptr,
    )
    if err:
        raise Krb5Error(context, err)

    creds.free_contents = 1

    return creds

TicketTimes = collections.namedtuple('TicketTimes', [
    'authtime',
    'starttime',
    'endtime',
    'renew_till',
])



# Support for Microsoft set/change password was added in MIT 1.7

import enum
import typing

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._creds cimport Creds
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    int32_t KRB5_KPASSWD_SUCCESS
    int32_t KRB5_KPASSWD_MALFORMED
    int32_t KRB5_KPASSWD_HARDERROR
    int32_t KRB5_KPASSWD_AUTHERROR
    int32_t KRB5_KPASSWD_SOFTERROR
    int32_t KRB5_KPASSWD_ACCESSDENIED
    int32_t KRB5_KPASSWD_BAD_VERSION
    int32_t KRB5_KPASSWD_INITIAL_FLAG_NEEDED

    krb5_error_code krb5_set_password(
        krb5_context context,
        krb5_creds *creds,
        const char *newpw,
        krb5_principal change_password_for,
        int *result_code,
        krb5_data *result_code_string,
        krb5_data *result_string
    ) nogil

    krb5_error_code krb5_set_password_using_ccache(
        krb5_context context,
        krb5_ccache ccache,
        const char *newpw,
        krb5_principal change_password_for,
        int *result_code,
        krb5_data *result_code_string,
        krb5_data *result_string
    ) nogil



class SetPasswordResultCode(enum.IntEnum):
    SUCCESS = KRB5_KPASSWD_SUCCESS
    MALFORMED = KRB5_KPASSWD_MALFORMED
    HARDERROR = KRB5_KPASSWD_HARDERROR
    AUTHERROR = KRB5_KPASSWD_AUTHERROR
    SOFTERROR = KRB5_KPASSWD_SOFTERROR
    ACCESSDENIED = KRB5_KPASSWD_ACCESSDENIED
    BAD_VERSION = KRB5_KPASSWD_BAD_VERSION
    INITIAL_FLAG_NEEDED = KRB5_KPASSWD_INITIAL_FLAG_NEEDED

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        if not isinstance(value, int):
            return None
        value = int(value)

        new_member = int.__new__(cls, value)
        new_member._name_ = f"Unknown_SetPasswordResultCode_{str(value).replace('-', 'm')}"
        new_member._value_ = value
        return cls._value2member_map_.setdefault(value, new_member)

class SetPasswordResult(typing.NamedTuple):
    result_code: SetPasswordResultCode
    result_code_string: bytes
    server_response: bytes

def set_password(
    Context context not None,
    Creds creds not None,
    const unsigned char[:] newpw not None,
    Principal change_password_for = None,
) -> SetPasswordResult:
    cdef krb5_error_code err = 0
    cdef int result_code
    cdef krb5_data krb5_result_code_string
    cdef krb5_data krb5_server_response
    cdef char *newpw_ptr
    cdef krb5_principal change_password_for_ptr = NULL
    cdef size_t length
    cdef char *value

    if len(newpw) > 0:
        newpw_ptr = <char *>&newpw[0]
    else:
        newpw_ptr = <char *>b""

    pykrb5_init_krb5_data(&krb5_result_code_string)
    pykrb5_init_krb5_data(&krb5_server_response)

    if change_password_for is not None:
        change_password_for_ptr = change_password_for.raw

    try:
        err = krb5_set_password(
            context.raw,
            creds.get_pointer(),
            newpw_ptr,
            change_password_for_ptr,
            &result_code,
            &krb5_result_code_string,
            &krb5_server_response
        )

        if err:
            raise Krb5Error(context, err)

        pykrb5_get_krb5_data(&krb5_result_code_string, &length, &value)

        if length == 0:
            result_code_string = b""
        else:
            result_code_string = <bytes>value[:length]

        pykrb5_get_krb5_data(&krb5_server_response, &length, &value)

        if length == 0:
            server_response = b""
        else:
            server_response = <bytes>value[:length]

        return SetPasswordResult(result_code, result_code_string, server_response)

    finally:
        pykrb5_free_data_contents(context.raw, &krb5_result_code_string)
        pykrb5_free_data_contents(context.raw, &krb5_server_response)

def set_password_using_ccache(
    Context context not None,
    CCache ccache not None,
    const unsigned char[:] newpw not None,
    Principal change_password_for = None,
) -> SetPasswordResult:
    cdef krb5_error_code err = 0
    cdef int result_code
    cdef krb5_data krb5_result_code_string
    cdef krb5_data krb5_server_response
    cdef char *newpw_ptr
    cdef krb5_principal change_password_for_ptr = NULL
    cdef size_t length
    cdef char *value

    if len(newpw) > 0:
        newpw_ptr = <char *>&newpw[0]
    else:
        newpw_ptr = <char *>b""

    pykrb5_init_krb5_data(&krb5_result_code_string)
    pykrb5_init_krb5_data(&krb5_server_response)

    if change_password_for is not None:
        change_password_for_ptr = change_password_for.raw

    try:
        err = krb5_set_password_using_ccache(
            context.raw,
            ccache.raw,
            newpw_ptr,
            change_password_for_ptr,
            &result_code,
            &krb5_result_code_string,
            &krb5_server_response
        )

        if err:
            raise Krb5Error(context, err)

        pykrb5_get_krb5_data(&krb5_result_code_string, &length, &value)

        if length == 0:
            result_code_string = b""
        else:
            result_code_string = <bytes>value[:length]

        pykrb5_get_krb5_data(&krb5_server_response, &length, &value)

        if length == 0:
            server_response = b""
        else:
            server_response = <bytes>value[:length]

        return SetPasswordResult(result_code, result_code_string, server_response)

    finally:
        pykrb5_free_data_contents(context.raw, &krb5_result_code_string)
        pykrb5_free_data_contents(context.raw, &krb5_server_response)




# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # Added in MIT 1.10
    krb5_boolean krb5_cc_support_switch(
        krb5_context context,
        const char *type,
    ) nogil


def cc_support_switch(
    Context context not None,
    const unsigned char[:] cache_type not None,
) -> bool:
    cdef const char *type_ptr = NULL
    if len(cache_type):
        type_ptr = <const char*>&cache_type[0]
    else:
        raise ValueError("cache_type cannot be an empty byte string")

    return bool(krb5_cc_support_switch(context.raw, type_ptr))



# Copyright: (c) 2024 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum
import typing

from krb5._ccache import CCache
from krb5._context import Context
from krb5._creds import Creds
from krb5._principal import Principal

class SetPasswordResultCode(enum.IntEnum):
    """Password change result constants returned by :meth:`set_password()`
    Follow RFC 3244 with additional Microsoft extensions.
    """

    SUCCESS = ...  # Success
    MALFORMED = ...  # Malformed request
    HARDERROR = ...  # Server error
    AUTHERROR = ...  # Authentication error
    SOFTERROR = ...  # Password change rejected
    ACCESSDENIED = ...  # Microsoft extension: Not authorized
    BAD_VERSION = ...  # Microsoft extension: Unknown RPC version
    INITIAL_FLAG_NEEDED = ...  # Microsoft extension:
    # The presented credentials were not obtained using a password directly

class SetPasswordResult(typing.NamedTuple):
    """The result returned by :meth:`set_password()` and
    :meth:`set_password_using_ccache()`.

    The `result_code` and `result_code_string` are the pure library responses.
    See `SetPasswordResultCode` for more information.

    The `server_response` is a server protocol message that may contain useful
    information about password policy violations or other errors.
    Despite RFC 3244, the server response is not standardized and may vary.
    Depending on `kpasswd` implementation, it may be returned as:\n
    - 30-byte binary Active Directory Policy Information
    - UTF-8 byte string (MIT KDC, potentially Heimdal KDC)
    - raw bytes (unknown or custom implementation)

    The trick is that Active Directory Policy Information always starts with
    `0x0000` signature to distinguish from UTF-8.
    So the client may try decoding the server response with either
    :meth:`ADPolicyInfo.from_bytes()` or :meth:`bytes.decode()`.
    And if the decoding fails with corresponding `ValueError` or
    `UnicodeDecodeError`, the raw bytes should be analyzed.

    See `ADPolicyInfo` for more information.
    """

    result_code: SetPasswordResultCode
    """The library result code of the password change operation."""
    result_code_string: bytes
    """The byte string representation of the result code."""
    server_response: bytes
    """Implementation-specific server response."""

def set_password(
    context: Context,
    creds: Creds,
    newpw: bytes,
    change_password_for: Principal | None = None,
) -> SetPasswordResult:
    """Set a password for a principal using specified credentials.


    This function implements the set password operation of ``RFC 3244``,
    for interoperability with Microsoft Windows implementations.
    It uses the credentials `creds` to set the password `newpw` for the
    principal `change_password_for`.
    If `change_password_for` is `None`, the password is set for the principal
    owning creds. If `change_password_for` is not `None`, the change is
    performed on the specified principal, assuming enough privileges.

    Note: the `creds` can be obtained using `get_init_creds_password()` with
    `in_tkt_service` set to ``kadmin/changepw``.

    Args:
        context: Krb5 context.
        creds: Credentials for kadmin/changepw service.
        newpw: New password.
        change_password_for: `None` or the principal to set the password for.

    Returns:
        SetPasswordResult: See `SetPasswordResult` for more information.
    """

def set_password_using_ccache(
    context: Context,
    ccache: CCache,
    newpw: bytes,
    change_password_for: Principal | None = None,
) -> SetPasswordResult:
    """Set a password for a principal using cached credentials.


    This function implements the set password operation of ``RFC 3244``,
    for interoperability with Microsoft Windows implementations.
    It uses the cached credentials from `ccache` to set the password `newpw` for
    the principal `change_password_for`.
    If `change_password_for` is `None`, the change is performed on the default
    principal in ccache. If `change_password_for` is not `None`, the change is
    performed on the specified principal.

    Note: the credentials can be obtained using `get_init_creds_password()` with
    `in_tkt_service` set to ``kadmin/changepw`` and stored to `ccache`.

    Args:
        context: Krb5 context.
        ccache: Credential cache.
        newpw: The new password.
        change_password_for: `None` or the principal to set the password for.

    Returns:
        SetPasswordResult: See `SetPasswordResult` for more information.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing

from libc.stdint cimport uintptr_t

from krb5._exceptions import Krb5Error
from krb5._principal import PrincipalParseFlags

from krb5._context cimport Context
from krb5._creds cimport Creds
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    krb5_error_code krb5_cc_close(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_default(
        krb5_context context,
        krb5_ccache *cache,
    ) nogil

    const char *krb5_cc_default_name(
        krb5_context context,
    ) nogil

    krb5_error_code krb5_cc_destroy(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_end_seq_get(
        krb5_context context,
        krb5_ccache cache,
        krb5_cc_cursor *cursor,
    ) nogil

    krb5_error_code krb5_cc_initialize(
        krb5_context context,
        krb5_ccache cache,
        krb5_principal principal,
    ) nogil

    const char *krb5_cc_get_name(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_get_principal(
        krb5_context context,
        krb5_ccache cache,
        krb5_principal *principal,
    ) nogil

    const char *krb5_cc_get_type(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_new_unique(
        krb5_context context,
        const char *type,
        const char *hint,
        krb5_ccache *id,
    ) nogil

    krb5_error_code krb5_cc_next_cred(
        krb5_context context,
        krb5_ccache cache,
        krb5_cc_cursor *cursor,
        krb5_creds *cred,
    ) nogil

    krb5_error_code krb5_cc_remove_cred(
        krb5_context context,
        krb5_ccache cache,
        int flags,
        krb5_creds *creds,
    ) nogil

    krb5_error_code krb5_cc_resolve(
        krb5_context context,
        const char *name,
        krb5_ccache *cache,
    ) nogil

    krb5_error_code krb5_cc_retrieve_cred(
        krb5_context context,
        krb5_ccache cache,
        int flags,
        krb5_creds *mcreds,
        krb5_creds *creds,
    ) nogil

    krb5_error_code krb5_cc_set_default_name(
        krb5_context context,
        const char *name,
    ) nogil

    krb5_error_code krb5_cc_store_cred(
        krb5_context context,
        krb5_ccache cache,
        krb5_creds *creds,
    ) nogil

    krb5_error_code krb5_cc_start_seq_get(
        krb5_context context,
        krb5_ccache cache,
        krb5_cc_cursor *cursor,
    ) nogil

    krb5_error_code krb5_cc_switch(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_get_config(
        krb5_context context,
        krb5_ccache id,
        krb5_const_principal principal,
        const char *key,
        krb5_data *data,
    );

    krb5_error_code krb5_cc_set_config(
        krb5_context context,
        krb5_ccache id,
        krb5_const_principal principal,
        const char *key,
        krb5_data *data,
    ) nogil

    int32_t KRB5_TC_MATCH_TIMES
    int32_t KRB5_TC_MATCH_IS_SKEY
    int32_t KRB5_TC_MATCH_FLAGS
    int32_t KRB5_TC_MATCH_TIMES_EXACT
    int32_t KRB5_TC_MATCH_FLAGS_EXACT
    int32_t KRB5_TC_MATCH_AUTHDATA
    int32_t KRB5_TC_MATCH_SRV_NAMEONLY
    int32_t KRB5_TC_MATCH_2ND_TKT
    int32_t KRB5_TC_MATCH_KTYPE
    int32_t KRB5_TC_SUPPORTED_KTYPES


_CredentialsRetrieveFlags_members = [
    ('none', 0),
    ('match_times', KRB5_TC_MATCH_TIMES),
    ('match_is_skey', KRB5_TC_MATCH_IS_SKEY),
    ('match_flags', KRB5_TC_MATCH_FLAGS),
    ('match_times_exact', KRB5_TC_MATCH_TIMES_EXACT),
    ('match_flags_exact', KRB5_TC_MATCH_FLAGS_EXACT),
    ('match_authdata', KRB5_TC_MATCH_AUTHDATA),
    ('match_srv_nameonly', KRB5_TC_MATCH_SRV_NAMEONLY),
    ('match_2nd_tkt', KRB5_TC_MATCH_2ND_TKT),
    ('match_keytype', KRB5_TC_MATCH_KTYPE),
]
# If KRB5_TC_SUPPORTED_KTYPES is not available it will be set to 0 in
# python_krb5.h
if KRB5_TC_SUPPORTED_KTYPES != 0:
    _CredentialsRetrieveFlags_members += [
        ('supported_ktypes', KRB5_TC_SUPPORTED_KTYPES),
    ]
CredentialsRetrieveFlags = enum.IntEnum('CredentialsRetrieveFlags', _CredentialsRetrieveFlags_members)


cdef class CCache:
    # cdef Context ctx
    # cdef krb5_ccache raw

    def __cinit__(CCache self, Context context):
        self.ctx = context
        self.raw = NULL

    def __dealloc__(CCache self):
        if self.raw:
            krb5_cc_close(self.ctx.raw, self.raw)
            self.raw = NULL

    def __iter__(CCache self) -> typing.Iterator[Creds]:
        cdef krb5_error_code err = 0
        cdef krb5_cc_cursor cursor

        if self.raw == NULL:
            return

        err = krb5_cc_start_seq_get(self.ctx.raw, self.raw, &cursor)
        if err:
            raise Krb5Error(self.ctx, err)

        try:
            while True:
                creds = Creds(self.ctx)
                err = krb5_cc_next_cred(
                    self.ctx.raw,
                    self.raw,
                    &cursor,
                    creds.get_pointer(),
                )
                if err:
                    break

                creds.free_contents = 1
                yield creds

        finally:
            err = krb5_cc_end_seq_get(self.ctx.raw, self.raw, &cursor)
            if err:
                raise Krb5Error(self.ctx, err)

    @property
    def addr(self) -> typing.Optional[int]:
        if self.raw:
            return <uintptr_t>self.raw

    @property
    def name(self) -> typing.Optional[bytes]:
        if self.raw:
            return cc_get_name(self.ctx, self)

    @property
    def principal(self) -> typing.Optional[Principal]:
        if self.raw:
            return cc_get_principal(self.ctx, self)

    @property
    def cache_type(self) -> typing.Optional[bytes]:
        if self.raw:
            return cc_get_type(self.ctx, self)

    def __repr__(CCache self) -> str:
        if self.raw:
            kwargs = [f"{k}={v}" for k, v in {
                'cache_type': self.cache_type.decode('utf-8'),
                'name': self.name.decode('utf-8'),
            }.items()]

            return f"CCache({', '.join(kwargs)})"

        else:
            return "CCache(NULL)"

    def __str__(CCache self) -> str:
        if self.raw:
            return f"{self.cache_type.decode('utf-8')}:{self.name.decode('utf-8')}"

        else:
            return "NULL"


def cc_default(
    Context context not None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    err = krb5_cc_default(context.raw, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_default_name(
    Context context not None,
) -> bytes:
    return <bytes>krb5_cc_default_name(context.raw)


def cc_destroy(
    Context context not None,
    CCache cache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_destroy(context.raw, cache.raw)
    if err:
        raise Krb5Error(context, err)
    cache.raw = NULL  # Stops dealloc from calling close


def cc_get_name(
    Context context not None,
    CCache cache not None,
) -> bytes:
    return krb5_cc_get_name(context.raw, cache.raw)


def cc_get_principal(
    Context context not None,
    CCache cache not None,
) -> Principal:
    princ = Principal(context, PrincipalParseFlags.none)
    cdef krb5_error_code err = 0

    err = krb5_cc_get_principal(context.raw, cache.raw, &princ.raw)
    if err:
        raise Krb5Error(context, err)

    return princ


def cc_get_type(
    Context context not None,
    CCache cache not None,
) -> bytes:
    return krb5_cc_get_type(context.raw, cache.raw)


def cc_initialize(
    Context context not None,
    CCache cache not None,
    Principal principal not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_initialize(context.raw, cache.raw, principal.raw)
    if err:
        raise Krb5Error(context, err)


def cc_new_unique(
    Context context not None,
    const unsigned char[:] cred_type not None,
    const unsigned char[:] hint = None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    # TODO: Test that this cannot be NULL
    if not len(cred_type):
        raise ValueError("cred_type must be set to a valid type value not an empty byte string")

    cdef const char *hint_ptr = NULL
    if hint is not None and len(hint):
        hint_ptr = <const char*>&hint[0]

    err = krb5_cc_new_unique(context.raw, <const char*>&cred_type[0], hint_ptr, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_remove_cred(
    Context context not None,
    CCache cache not None,
    int flags,
    Creds creds not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_remove_cred(
        context.raw,
        cache.raw,
        flags,
        creds.get_pointer(),
    )
    if err:
        raise Krb5Error(context, err)


def cc_resolve(
    Context context not None,
    const unsigned char[:] name not None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if len(name):
        name_ptr = <const char*>&name[0]
    else:
        raise ValueError("name cannot be an empty byte string")

    err = krb5_cc_resolve(context.raw, name_ptr, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_retrieve_cred(
    Context context not None,
    CCache cache not None,
    int flags,
    Creds mcreds not None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    err = krb5_cc_retrieve_cred(
        context.raw,
        cache.raw,
        flags,
        mcreds.get_pointer(),
        creds.get_pointer())
    if err:
        raise Krb5Error(context, err)

    creds.free_contents = 1

    return creds


def cc_set_default_name(
    Context context not None,
    const unsigned char[:] name,
) -> None:
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if name is not None and len(name):
        name_ptr = <const char *>&name[0]

    err = krb5_cc_set_default_name(context.raw, name_ptr)
    if err:
        raise Krb5Error(context, err)


def cc_store_cred(
    Context context not None,
    CCache cache not None,
    Creds creds not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_store_cred(context.raw, cache.raw, creds.get_pointer())
    if err:
        raise Krb5Error(context, err)


def cc_switch(
    Context context not None,
    CCache cache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_switch(context.raw, cache.raw)
    if err:
        raise Krb5Error(context, err)


def cc_get_config(
    Context context not None,
    CCache cache not None,
    Principal principal,
    const unsigned char[:] key not None,
) -> bytes:
    cdef krb5_error_code err = 0

    cdef krb5_principal principal_raw = NULL
    if principal is not None:
        principal_raw = principal.raw

    cdef const char *key_ptr
    if len(key):
        key_ptr = <const char*>&key[0]
    else:
        key_ptr = ""

    cdef krb5_data data
    err = krb5_cc_get_config(context.raw, cache.raw, principal_raw, key_ptr, &data)
    if err:
        raise Krb5Error(context, err)

    cdef size_t length
    cdef char *value
    pykrb5_get_krb5_data(&data, &length, &value)

    if length == 0:
        data_bytes = b""
    else:
        data_bytes = <bytes>value[:length]

    pykrb5_free_data_contents(context.raw, &data)

    return data_bytes


def cc_set_config(
    Context context not None,
    CCache cache not None,
    Principal principal,
    const unsigned char[:] key not None,
    const unsigned char[:] data,
) -> None:
    cdef krb5_error_code err = 0

    cdef krb5_principal principal_raw = NULL
    if principal is not None:
        principal_raw = principal.raw

    cdef const char *key_ptr
    if len(key):
        key_ptr = <const char*>&key[0]
    else:
        key_ptr = ""

    cdef krb5_data data_raw
    cdef krb5_data* data_ptr
    if data is None:
        data_ptr = NULL
    else:
        if len(data) == 0:
            pykrb5_set_krb5_data(&data_raw, 0, "")
        else:
            pykrb5_set_krb5_data(&data_raw, len(data), <char *>&data[0])
        data_ptr = &data_raw

    err = krb5_cc_set_config(context.raw, cache.raw, principal_raw, key_ptr, data_ptr)
    if err:
        raise Krb5Error(context, err)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef class Creds:
    cdef Context ctx
    cdef int free_contents
    cdef krb5_creds* _raw
    cdef int _free_raw

    cdef void* set_raw_from_lib(Creds self, krb5_creds* raw)
    cdef krb5_creds *get_pointer(Creds self)


cdef class InitCredsContext:
    cdef Context ctx
    cdef krb5_init_creds_context raw


cdef class Krb5Prompt:
    pass



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

# These APIs were added in MIT 1.20, to be compat with 1.18 we need to define
# them separately.

import collections
import typing

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._creds cimport Creds
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    krb5_error_code krb5_marshal_credentials(
        krb5_context context,
        krb5_creds *creds,
        krb5_data **data
    ) nogil

    krb5_error_code krb5_unmarshal_credentials(
        krb5_context context,
        krb5_data *data,
        krb5_creds **creds,
    ) nogil

    void krb5_free_data(
        krb5_context context,
        krb5_data *val,
    ) nogil

def marshal_credentials(
    Context context not None,
    Creds creds not None,
) -> bytes:
    cdef krb5_error_code err = 0
    cdef krb5_data *data = NULL
    cdef size_t length
    cdef char *value

    try:
        err = krb5_marshal_credentials(context.raw, creds.get_pointer(), &data)

        if err:
            raise Krb5Error(context, err)

        pykrb5_get_krb5_data(data, &length, &value)

        if length == 0:
            data_bytes = b""
        else:
            data_bytes = value[:length]

        return data_bytes

    finally:
        if NULL != data:
            krb5_free_data(context.raw, data)

def unmarshal_credentials(
    Context context not None,
    const unsigned char[:] data not None,
) -> Creds:
    cdef krb5_error_code err = 0
    cdef krb5_creds* raw_creds = NULL
    cdef krb5_data data_raw

    if len(data) == 0:
        pykrb5_set_krb5_data(&data_raw, 0, "")
    else:
        pykrb5_set_krb5_data(&data_raw, len(data), <char *>&data[0])

    err = krb5_unmarshal_credentials(context.raw, &data_raw, &raw_creds)
    if err:
        raise Krb5Error(context, err)

    creds = Creds(context)
    creds.set_raw_from_lib(raw_creds)

    return creds



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_get_init_creds_opt_set_fast_ccache(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        krb5_ccache ccache,
    ) nogil

    krb5_error_code krb5_get_init_creds_opt_set_fast_flags(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        krb5_int32 flags,
    ) nogil

    krb5_error_code krb5_get_init_creds_opt_set_fast_ccache_name(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        const char *fast_ccache_name,
    ) nogil

    krb5_error_code krb5_get_init_creds_opt_set_out_ccache(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        krb5_ccache ccache,
    ) nogil

    krb5_error_code krb5_get_init_creds_opt_set_pa(
        krb5_context context,
        krb5_get_init_creds_opt *opt,
        const char *attr,
        const char *value,
    ) nogil

    int32_t KRB5_FAST_REQUIRED


class FastFlags(enum.IntEnum):
    none = 0
    required = KRB5_FAST_REQUIRED


def get_init_creds_opt_set_fast_ccache(
    Context context not None,
    GetInitCredsOpt opt not None,
    CCache ccache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_get_init_creds_opt_set_fast_ccache(context.raw, opt.raw, ccache.raw)
    if err:
        raise Krb5Error(context, err)


def get_init_creds_opt_set_fast_flags(
    Context context not None,
    GetInitCredsOpt opt not None,
    krb5_int32 flags,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_get_init_creds_opt_set_fast_flags(context.raw, opt.raw, flags)
    if err:
        raise Krb5Error(context, err)


def get_init_creds_opt_set_fast_ccache_name(
    Context context not None,
    GetInitCredsOpt opt not None,
    const unsigned char[:] name not None,
) -> None:
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if len(name):
        name_ptr = <const char*>&name[0]
    else:
        raise ValueError("CCache name must be set")

    err = krb5_get_init_creds_opt_set_fast_ccache_name(context.raw, opt.raw, name_ptr)
    if err:
        raise Krb5Error(context, err)


def get_init_creds_opt_set_out_ccache(
    Context context not None,
    GetInitCredsOpt opt not None,
    CCache ccache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_get_init_creds_opt_set_out_ccache(context.raw, opt.raw, ccache.raw)
    if err:
        raise Krb5Error(context, err)


def get_init_creds_opt_set_pa(
    Context context not None,
    GetInitCredsOpt opt not None,
    const unsigned char[:] attr not None,
    const unsigned char[:] value not None,
) -> None:
    cdef krb5_error_code err = 0

    cdef char *attr_ptr
    if len(attr):
        attr_ptr = <char *>&attr[0]
    else:
        raise ValueError("attr must be set")

    cdef char *value_ptr
    if len(value):
        value_ptr = <char *>&value[0]
    else:
        raise ValueError("value must be set")

    err = krb5_get_init_creds_opt_set_pa(context.raw, opt.raw, attr_ptr, value_ptr)
    if err:
        raise Krb5Error(context, err)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pathlib
import time
import typing

import k5test
import pytest

import krb5


class MockPrompt(krb5.Krb5Prompt):
    def __init__(self, responses: typing.List[bytes]) -> None:
        self.init_calls: typing.List[typing.Tuple[typing.Optional[bytes], typing.Optional[bytes], int]] = []
        self.prompt_calls: typing.List[typing.Tuple[bytes, bool]] = []
        self._responses = responses

    def init(
        self,
        name: typing.Optional[bytes],
        banner: typing.Optional[bytes],
        num_prompts: int,
    ) -> None:
        self.init_calls.append((name, banner, num_prompts))

    def prompt(self, msg: bytes, hidden: bool) -> bytes:
        self.prompt_calls.append((msg, hidden))
        return self._responses.pop(0)


def test_TicketFlags() -> None:
    # proxy (1 << 4) and two unknown flags (1 << 24 and 1 << 31) set
    flags = krb5.TicketFlags(0x81000010)

    assert krb5.TicketFlags.proxy in flags
    assert krb5.TicketFlags.forwarded not in flags

    # Clear proxy and forwarded, leave unknown flags intact
    flags = flags & ~(krb5.TicketFlags.proxy | krb5.TicketFlags.forwarded)
    assert krb5.TicketFlags.proxy not in flags
    assert krb5.TicketFlags.forwarded not in flags
    assert flags == 0x81000000
    assert type(flags) == krb5.TicketFlags

    flags = flags | krb5.TicketFlags.postdated
    assert krb5.TicketFlags.postdated in flags
    assert flags == 0x81000040
    assert type(flags) == krb5.TicketFlags


def test_get_init_creds_keytab(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.host_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    kt = krb5.kt_default(ctx)

    creds = krb5.get_init_creds_keytab(ctx, princ, opt, kt)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"

    assert creds.client.name == realm.host_princ.encode()
    assert creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    assert len(creds.keyblock.data) > 0
    assert str(creds.times).startswith("TicketTimes(authtime=")
    assert krb5.TicketFlags.initial in creds.ticket_flags
    # creds.addresses
    assert len(creds.ticket) > 0
    assert creds.second_ticket == b""
    # creds.authdata


def test_get_init_creds_password(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)

    creds = krb5.get_init_creds_password(ctx, princ, opt, realm.password("user").encode())
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"


def test_get_init_creds_password_prompt(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    prompter = MockPrompt([realm.password("user").encode()])

    creds = krb5.get_init_creds_password(ctx, princ, opt, prompter=prompter)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"

    assert len(prompter.init_calls) == 1
    assert prompter.init_calls[0] == (None, None, 1)
    assert len(prompter.prompt_calls) == 1

    expected = (
        f"Password for {realm.user_princ}" if realm.provider == "mit" else f"{realm.user_princ}'s Password: "
    ).encode()
    assert prompter.prompt_calls[0] == (expected, True)


def test_get_init_creds_password_prompt_failure(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    prompter = MockPrompt([])

    with pytest.raises(krb5.Krb5Error):
        krb5.get_init_creds_password(ctx, princ, opt, prompter=prompter)

    assert len(prompter.init_calls) == 1
    assert prompter.init_calls[0] == (None, None, 1)
    assert len(prompter.prompt_calls) == 1


def test_get_creds_keytab(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.host_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    kt = krb5.kt_resolve(ctx, realm.keytab.encode())
    krb5.init_creds_set_keytab(ctx, creds_ctx, kt)
    krb5.init_creds_get(ctx, creds_ctx)

    creds = krb5.init_creds_get_creds(ctx, creds_ctx)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"


def test_get_creds_keytab_wrong_principal(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    kt = krb5.kt_resolve(ctx, realm.keytab.encode())

    expected = "no suitable keys for" if realm.provider == "mit" else "Failed to find"
    with pytest.raises(krb5.Krb5Error, match=expected):
        krb5.init_creds_set_keytab(ctx, creds_ctx, kt)


def test_init_creds_set_password(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    krb5.init_creds_set_password(ctx, creds_ctx, realm.password("user").encode())
    krb5.init_creds_get(ctx, creds_ctx)

    creds = krb5.init_creds_get_creds(ctx, creds_ctx)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"


def test_init_creds_set_password_invalid(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    krb5.init_creds_set_password(ctx, creds_ctx, b"invalid")

    # Too many different error messages - just expect an error
    with pytest.raises(krb5.Krb5Error):
        krb5.init_creds_get(ctx, creds_ctx)


def test_renew_creds(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    # Ask for a renewable ticket
    krb5.get_init_creds_opt_set_renew_life(opt, 1024)
    creds = krb5.get_init_creds_password(ctx, princ, opt, realm.password("user").encode())

    assert creds.client.name == realm.user_princ.encode()
    assert creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    assert krb5.TicketFlags.initial in creds.ticket_flags
    assert krb5.TicketFlags.renewable in creds.ticket_flags

    flags_raw = creds.ticket_flags_raw
    flags_raw_reversed = 0
    for i in range(32):
        if flags_raw & (1 << i):
            flags_raw_reversed |= 1 << (31 - i)
    if realm.provider.lower() == "heimdal":
        assert creds.ticket_flags == flags_raw
    else:
        assert creds.ticket_flags == flags_raw_reversed

    cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.cc_initialize(ctx, cc, princ)
    krb5.cc_store_cred(ctx, cc, creds)

    new_creds = krb5.get_renewed_creds(ctx, creds.client, cc)
    assert new_creds.client.name == realm.user_princ.encode()
    assert new_creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    if realm.provider.lower() == "heimdal":
        # The MIT KDC seems to return renewed tickets with the 'initial' flag
        # set.
        assert krb5.TicketFlags.initial not in new_creds.ticket_flags

    new_creds = krb5.get_renewed_creds(ctx, creds.client, cc, b"krbtgt/KRBTEST.COM@KRBTEST.COM")
    assert new_creds.client.name == realm.user_princ.encode()
    assert new_creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    if realm.provider.lower() == "heimdal":
        assert krb5.TicketFlags.initial not in new_creds.ticket_flags


@pytest.mark.requires_api("get_validated_creds")
def test_validate_creds(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    # Get postdated ticket, ticket will be valid after 1s
    creds = krb5.get_init_creds_password(ctx, princ, opt, realm.password("user").encode(), start_time=1)

    assert creds.client.name == realm.user_princ.encode()
    assert creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    # Ticket flags for creds should have TKT_FLG_POSTDATED and TKT_FLG_INVALID set
    assert krb5.TicketFlags.postdated in creds.ticket_flags
    assert krb5.TicketFlags.invalid in creds.ticket_flags

    cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.cc_initialize(ctx, cc, princ)
    krb5.cc_store_cred(ctx, cc, creds)

    start_time = time.time()
    while True:
        try:
            new_creds = krb5.get_validated_creds(ctx, creds.client, cc)
            break
        except krb5.Krb5Error as e:
            # Retry within the first 5s when the error is
            # KRB5KRB_AP_ERR_TKT_NYV ("Ticket not yet valid"). The ticket should
            # normally be valid after 0-1s.
            if (time.time() - start_time < 5) and e.err_code == -1765328351:
                # Retry
                time.sleep(0.1)
            else:
                raise
    assert new_creds.client.name == realm.user_princ.encode()
    assert new_creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    # Ticket flags for new_creds should have TKT_FLG_POSTDATED set and TKT_FLG_INVALID cleared
    assert krb5.TicketFlags.postdated in new_creds.ticket_flags
    assert krb5.TicketFlags.invalid not in new_creds.ticket_flags

    new_creds = krb5.get_validated_creds(ctx, creds.client, cc, b"krbtgt/KRBTEST.COM@KRBTEST.COM")
    assert new_creds.client.name == realm.user_princ.encode()
    assert new_creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    # Ticket flags for new_creds should have TKT_FLG_POSTDATED set and TKT_FLG_INVALID cleared
    assert krb5.TicketFlags.postdated in new_creds.ticket_flags
    assert krb5.TicketFlags.invalid not in new_creds.ticket_flags


@pytest.mark.requires_api("get_etype_info")
def test_get_etype_info(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())

    # Get information about etype, salt and s2kparams for the principal
    etype, salt, s2kparams = krb5.get_etype_info(ctx, princ)

    # Check whether the KDC returned the etype-info
    if salt is None:
        raise Exception("Failed to get etype-info")

    # Use the string-to-key function to get the user's key
    kb = krb5.c_string_to_key(ctx, etype, realm.password("user").encode(), salt, s2kparams)

    # Write the key into a keytab
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, kb)

    # Get credentials using this keytab
    opt = krb5.get_init_creds_opt_alloc(ctx)
    creds = krb5.get_init_creds_keytab(ctx, princ, opt, kt)

    # Verify the credentials
    assert creds.client.name == realm.user_princ.encode()
    assert creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"


@pytest.mark.requires_api("marshal_credentials")
def test_creds_serialization(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    creds = krb5.get_init_creds_password(ctx, princ, opt, realm.password("user").encode())
    assert isinstance(creds, krb5.Creds)

    with pytest.raises(krb5.Krb5Error):
        krb5.unmarshal_credentials(ctx, b"invalid")

    with pytest.raises(krb5.Krb5Error):
        krb5.unmarshal_credentials(ctx, b"")

    marshalled_actual = krb5.marshal_credentials(ctx, creds)
    assert isinstance(marshalled_actual, bytes)
    assert len(marshalled_actual) > 0

    uncreds = krb5.unmarshal_credentials(ctx, marshalled_actual)
    assert isinstance(uncreds, krb5.Creds)
    assert str(uncreds) == "Creds"

    assert id(creds) != id(uncreds)
    assert creds.client.name == uncreds.client.name
    assert creds.ticket == uncreds.ticket
    assert creds.keyblock.data == uncreds.keyblock.data
    assert creds.times.endtime == uncreds.times.endtime



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context

class GetInitCredsOpt:
    """Kerberos Initial Credential Options object.

    This class represents an initial credential options object.

    Args:
        context: Krb5 context.
    """

def get_init_creds_opt_alloc(
    context: Context,
) -> GetInitCredsOpt:
    """Allocate a new initial credential options object.

    Creates a new options structure that control how credential are gotten.

    Args:
        context: Krb5 context.

    Returns:
        GetInitCredsOpt: The initial credential options object.
    """

def get_init_creds_opt_set_anonymous(
    opt: GetInitCredsOpt,
    anonymous: bool,
) -> None:
    """Set anonymous details in the initial credential options.

    This function may be used to request anonymous credentials from the KDC.

    Note:
        Anonymous credentials are only a request; clients must verify that
        credentials are anonymous if that is a requirement.

    Args:
        opt: The initial credential options.
        anonymous: Whether to set or unset the anonymous option.
    """

def get_init_creds_opt_set_canonicalize(
    opt: GetInitCredsOpt,
    canonicalize: bool,
) -> None:
    """Set canonicalization details in the initial credential options.

    Sets whether the credential client principal should be canonicalized by
    the KDC.

    Args:
        opt: The initial credential options.
        canonicalize: Whether to set or unset the canonicalize option.
    """

def get_init_creds_opt_set_etype_list(
    opt: GetInitCredsOpt,
    etypes: typing.Iterable[int],
) -> None:
    """Set allowable encryptions types in the initial credential options.

    Sets the allowable encryption types in the initial credential options. Use
    :meth:`string_to_enctype` to convert an encryption type string to the int
    identifier.

    Args:
        opt: The initial credential options.
        etypes: The list of allowable encryption types.
    """

def get_init_creds_opt_set_forwardable(
    opt: GetInitCredsOpt,
    forwardable: bool,
) -> None:
    """Set forwardable details in the initial credential options.

    Sets whether the credentials are to be forwardable or not.

    Args:
        opt: The initial credential options.
        forwardable: Whether to set or unset the forwardable option.
    """

def get_init_creds_opt_set_proxiable(
    opt: GetInitCredsOpt,
    proxiable: bool,
) -> None:
    """Set proxiable details in the initial credential options.

    Sets whether the credentials are to be proxiable or not.

    Args:
        opt: The initial credential options.
        proxiable: Whether to set or unset the proxiable option.
    """

def get_init_creds_opt_set_renew_life(
    opt: GetInitCredsOpt,
    renew_life: int,
) -> None:
    """Set the ticket renewal lifetime.

    Sets the ticket renewal lifetime in seconds in the initial credential
    options.

    Args:
        opt: The initial credential options.
        renew_life: The ticket renewal lifetime in seconds.
    """

def get_init_creds_opt_set_salt(
    opt: GetInitCredsOpt,
    salt: bytes,
) -> None:
    """Set salt for optimistic preauth in initial credential options.

    When getting initial credentials with a password, a salt string it used to
    convert the password to a key. Normally this salt is obtained from the
    first KDC reply, but when performing optimistic preauthentication, the
    client may need to supply the salt string with this function.

    Args:
        opt: The initial credential options.
        salt: The salt to set.
    """

def get_init_creds_opt_set_tkt_life(
    opt: GetInitCredsOpt,
    tkt_life: int,
) -> None:
    """Set the ticket lifetime.

    Sets the ticket lifetime in seconds in the initial credential options.

    Args:
        opt: The initial credential options.
        tkt_life: The ticket lifetime in seconds.
    """



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pytest

import krb5


def test_init_keyblock_empty() -> None:
    ctx = krb5.init_context()
    kb = krb5.init_keyblock(ctx, 0, None)

    assert len(kb) == 0
    assert kb.enctype == 0
    assert kb.data == b""
    assert str(kb) == "KeyBlock 0"
    assert repr(kb) == "KeyBlock(enctype=0, length=0)"


def test_init_keyblock_data() -> None:
    ctx = krb5.init_context()
    kb = krb5.init_keyblock(ctx, 17, b"\xff" * 16)

    assert len(kb) == 16
    assert kb.enctype == 17
    assert kb.data == b"\xff" * 16
    assert str(kb) == "KeyBlock 17"
    assert repr(kb) == "KeyBlock(enctype=17, length=16)"


@pytest.mark.requires_api("c_string_to_key")
def test_c_string_to_key() -> None:
    ctx = krb5.init_context()

    salt = b"EXAMPLE.COMtestuser"
    password = b"Some Password"

    kb = krb5.c_string_to_key(ctx, 17, salt, password)
    assert kb.enctype == 17
    assert len(kb) == 16
    assert kb.data == b"\xd2\x153\xcd\xd9\x7fR\xe6\x11U]7\xac\x12[\xf6"

    kb = krb5.c_string_to_key(ctx, 17, salt, password, b"\x00\x00\x10\x00")
    assert kb.enctype == 17
    assert len(kb) == 16
    assert kb.data == b"\xd2\x153\xcd\xd9\x7fR\xe6\x11U]7\xac\x12[\xf6"

    kb = krb5.c_string_to_key(ctx, 18, salt, password, b"\x00\x00\x10\x00")
    assert kb.enctype == 18
    assert len(kb) == 32
    assert kb.data == b"\x0b\x12\x05\xb0\xc4\xe7\x0e\xf1\xbf\xf5\xeaJ\x1a\x80?~@m\x7f\xcakPk\x08\xa6\x99\x15\xd6s\r&("

    kb = krb5.c_string_to_key(ctx, 18, salt, password, b"\x00\x01\x23\x45")
    assert kb.enctype == 18
    assert len(kb) == 32
    assert (
        kb.data
        == b"\x11\xcc\x10\x0e\xff$\xc1SL^d\x00\xe2\x83\x08\xefxM\x12\x92\x18:\x1c\x9b\xd2w\xf5\xfd\xb9\x13\xe5\xd1"
    )



import typing

from krb5._exceptions import Krb5Error

from libc.string cimport strlen

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_chpw_message(
        krb5_context context,
        const krb5_data *server_string,
        char **message_out
    ) nogil

    void krb5_free_string(
        krb5_context context,
        char *string
    ) nogil

def chpw_message(
    Context context not None,
    const unsigned char[:] server_string not None,
) -> bytes:
    cdef krb5_error_code err = 0
    cdef krb5_data server_string_raw
    cdef char *message_out = NULL

    try:
        if len(server_string) == 0:
            pykrb5_set_krb5_data(&server_string_raw, 0, "")
        else:
            pykrb5_set_krb5_data(&server_string_raw, len(server_string), <char *>&server_string[0])

        err = krb5_chpw_message(context.raw, &server_string_raw, &message_out)

        if err:
            raise Krb5Error(context, err)

        if message_out is NULL:
            return b""
        else:
            message_len = strlen(message_out)
            return <bytes>message_out[:message_len]

    finally:
        krb5_free_string(context.raw, message_out)


import locale

import k5test
import pytest

import krb5


@pytest.mark.requires_api("chpw_message")
def test_chpw_message() -> None:

    # MIT Kerberos test samples, https://github.com/krb5/krb5, human readable
    adpi_tests = {
        "complex": [
            (
                b"\0\0",
                b"\0\0\0\0",
                b"\0\0\0\0",
                b"\0\0\0\1",
                b"\0\0\0\0\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
            ),
            (
                "The password must include numbers or symbols.",
                "Don't include any part of your name in the password.",
            ),
        ],
        "length": [
            (
                b"\0\0",
                b"\0\0\0\x0d",
                b"\0\0\0\0",
                b"\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
            ),
            ("The password must contain at least 13 characters.",),
        ],
        "history": [
            (
                b"\0\0",
                b"\0\0\0\0",
                b"\0\0\0\x09",
                b"\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
            ),
            ("The password must be different from the previous 9 passwords."),
        ],
        "age": [
            (
                b"\0\0",
                b"\0\0\0\0",
                b"\0\0\0\0",
                b"\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
                b"\0\0\x01\x92\x54\xd3\x80\0",
            ),
            ("The password can only be changed every 2 days."),
        ],
        "combined": [
            (
                b"\0\0",
                b"\0\0\0\x05",
                b"\0\0\0\x0d",
                b"\0\0\0\x01",
                b"\0\0\0\0\0\0\0\0",
                b"\0\0\0\xc9\x2a\x69\xc0\0",
            ),
            (
                "The password can only be changed once a day.",
                "The password must be different from the previous 13 passwords.",
                "The password must contain at least 5 characters.",
                "The password must include numbers or symbols.",
                "Don't include any part of your name in the password.",
            ),
        ],
        "unknown": [
            (
                b"\0\0",
                b"\0\0\0\0",
                b"\0\0\0\0",
                b"\x80\0\0\1",
                b"\0\0\0\0\0\0\0\0",
                b"\0\0\0\0\0\0\0\0",
            )
        ],
    }

    locale.setlocale(locale.LC_ALL, "C")

    ctx = krb5.init_context()

    samples = {k: b"".join(v[0]) if isinstance(v, list) else b"" for k, v in adpi_tests.items()}

    for k, test in adpi_tests.items():
        if isinstance(test, list) and len(test) > 1:
            phrases = test[1]
            for phrase in phrases:
                message = krb5.chpw_message(ctx, samples[k])
                assert message.decode().find(phrase) >= 0

    assert krb5.ADPolicyInfoProp.COMPLEX in krb5.ADPolicyInfo.from_bytes(samples["complex"]).properties
    assert krb5.ADPolicyInfo.from_bytes(samples["length"]).min_length == 13
    assert krb5.ADPolicyInfo.from_bytes(samples["history"]).history == 9
    assert krb5.ADPolicyInfo.from_bytes(samples["age"]).min_age == 2 * 86400 * 10_000_000
    assert krb5.ADPolicyInfo.from_bytes(samples["combined"]).min_length == 5
    assert krb5.ADPolicyInfo.from_bytes(samples["combined"]).history == 13
    assert krb5.ADPolicyInfo.from_bytes(samples["combined"]).min_age == 1 * 86400 * 10_000_000
    assert krb5.ADPolicyInfoProp.COMPLEX in krb5.ADPolicyInfo.from_bytes(samples["combined"]).properties
    assert 0x80000000 & krb5.ADPolicyInfo.from_bytes(samples["unknown"]).properties != 0


def test_set_password(realm: k5test.K5Realm) -> None:

    if realm.provider != "mit":
        # Heimdal testing requires complicated kadmind and kpasswdd setup
        return

    realm.start_kadmind()

    princ_name = "exp@" + realm.realm
    old_password = realm.password("userexp")
    weak_password = "sh0rt"
    empty_password = ""
    new_password = realm.password("user")
    new_password2 = realm.password("user~")

    # setup
    realm.run_kadminl(["addpol", "-minlength", "6", "-minclasses", "2", "pwpol"])
    realm.run_kadminl(["addprinc", "-pw", old_password, "-policy", "pwpol", "+needchange", princ_name])

    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, princ_name.encode())
    admin_princ = krb5.parse_name_flags(ctx, realm.admin_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)

    # admin creds; will be reused with ccache as well
    admin_creds = krb5.get_init_creds_password(
        ctx, admin_princ, opt, realm.password("admin").encode(), in_tkt_service=b"kadmin/changepw"
    )
    assert isinstance(admin_creds, krb5.Creds)

    # set_password for creds owner (self)
    with pytest.raises(krb5.Krb5Error) as exc:
        krb5.get_init_creds_password(ctx, princ, opt, password=old_password.encode())
    assert exc.value.err_code == -1765328361  # KRB5KDC_ERR_KEY_EXP

    creds = krb5.get_init_creds_password(ctx, princ, opt, old_password.encode(), in_tkt_service=b"kadmin/changepw")
    assert isinstance(creds, krb5.Creds)

    (result_code, result_code_string, server_response) = krb5.set_password(ctx, creds, empty_password.encode())
    assert result_code == krb5.SetPasswordResultCode.SOFTERROR
    assert result_code_string.find(b"rejected") > 0
    assert server_response.find(b"too short") > 0

    (result_code, result_code_string, server_response) = krb5.set_password(ctx, creds, weak_password.encode())
    assert result_code == krb5.SetPasswordResultCode.SOFTERROR
    assert result_code_string.find(b"rejected") > 0
    assert server_response.find(b"too short") > 0

    (result_code, result_code_string, server_response) = krb5.set_password(ctx, creds, new_password.encode())
    assert result_code == krb5.SetPasswordResultCode.SUCCESS

    creds = krb5.get_init_creds_password(ctx, princ, opt, new_password.encode())
    assert isinstance(creds, krb5.Creds)
    assert creds.client.name == princ.name

    # set_password for other principal using admin creds
    (result_code, result_code_string, server_response) = krb5.set_password(
        ctx, admin_creds, new_password2.encode(), change_password_for=princ
    )
    assert result_code == krb5.SetPasswordResultCode.SUCCESS

    creds = krb5.get_init_creds_password(ctx, princ, opt, new_password2.encode())
    assert isinstance(creds, krb5.Creds)
    assert creds.client.name == princ.name

    # reset password locally for next test
    realm.run_kadminl(["cpw", "-pw", old_password, princ_name])
    realm.run_kadminl(["modprinc", "+needchange", princ_name])

    # set_password_using_ccache
    with pytest.raises(krb5.Krb5Error) as exc:
        krb5.get_init_creds_password(ctx, princ, opt, password=old_password.encode())
    assert exc.value.err_code == -1765328361  # KRB5KDC_ERR_KEY_EXP

    creds = krb5.get_init_creds_password(ctx, princ, opt, old_password.encode(), in_tkt_service=b"kadmin/changepw")
    assert isinstance(creds, krb5.Creds)
    assert creds.client.name == princ.name

    cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.cc_initialize(ctx, cc, princ)
    krb5.cc_store_cred(ctx, cc, creds)

    (result_code, result_code_string, server_response) = krb5.set_password_using_ccache(
        ctx, cc, empty_password.encode(), princ
    )
    assert result_code == krb5.SetPasswordResultCode.SOFTERROR
    assert result_code_string.find(b"rejected") > 0
    assert server_response.find(b"too short") > 0

    (result_code, result_code_string, server_response) = krb5.set_password_using_ccache(
        ctx, cc, weak_password.encode(), princ
    )
    assert result_code == krb5.SetPasswordResultCode.SOFTERROR
    assert result_code_string.find(b"rejected") > 0
    assert server_response.find(b"too short") > 0

    (result_code, result_code_string, server_response) = krb5.set_password_using_ccache(
        ctx, cc, new_password.encode(), princ
    )
    assert result_code == krb5.SetPasswordResultCode.SUCCESS

    creds = krb5.get_init_creds_password(ctx, princ, opt, new_password.encode())
    assert isinstance(creds, krb5.Creds)
    assert creds.client.name == princ.name

    krb5.cc_destroy(ctx, cc)

    admin_cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.cc_initialize(ctx, admin_cc, admin_princ)
    krb5.cc_store_cred(ctx, admin_cc, admin_creds)

    # set_password for other principal using admin ccache
    (result_code, result_code_string, server_response) = krb5.set_password_using_ccache(
        ctx, admin_cc, new_password2.encode(), change_password_for=princ
    )
    assert result_code == krb5.SetPasswordResultCode.SUCCESS

    creds = krb5.get_init_creds_password(ctx, princ, opt, new_password2.encode())
    assert isinstance(creds, krb5.Creds)
    assert creds.client.name == princ.name

    realm.run_kadminl(["delprinc", "-force", princ_name])
    realm.run_kadminl(["delpol", "-force", "pwpol"])

    krb5.cc_destroy(ctx, admin_cc)

    realm.stop_kadmind()



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._ccache import CCache
from krb5._context import Context
from krb5._creds_opt import GetInitCredsOpt

def get_init_creds_opt_set_in_ccache(
    context: Context,
    opt: GetInitCredsOpt,
    ccache: CCache,
) -> None:
    """Set an input credential cache in initial credential options.

    If an input credential cache is set, then the krb5_get_init_creds family of
    APIs will read settings from it. Setting an input ccache is desirable when
    the application wishes to perform authentication in the same way (using the
    same preauthentication mechanisms, and making the same non-security
    sensitive choices) as the previous authentication attempt, which stored
    information in the passed-in cache.

    Args:
        context: Krb5 context.
        opt: The initial credential options.
        ccache: The credential cache to set.
    """
